﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Vogel.GeoTools;
using ArcViewShapeFileDLL;
using System.Windows.Forms;
using System.Drawing;
using SharpKml;
using Vogel.GeoTools;

namespace Vogel.AFS2
{
    // **************************** TOC Files ***************
    public class TOCFile
    {
        public Cultivation cultivation { get; set; }
        public string Name { get; set; }
        public string FullName { get; set; }

        public TOCFileInfo Info { get; set; }

        public TOCFile(string TOCpath)
        {
            this.FullName = TOCpath;
            var info = new System.IO.FileInfo(TOCpath);
            Name = info.Name;
            cultivation = new Cultivation();
        }

        public TOCFile()
        {
            cultivation = new Cultivation();
            Info = new TOCFileInfo();
        }

        public void LoadFromShapefile(string shpFilename)
        {
            System.IO.FileInfo info = new System.IO.FileInfo(shpFilename);
            this.Name = info.Name.Substring(0, info.Name.Length - 4);
            this.FullName = info.FullName.Substring(0, info.FullName.Length - 4) + ".toc";

            int nb_error = 0;

            ShapeFiles MyShape = MyShape = new ShapeFiles();
            try
            {
                MyShape.OpenShape(shpFilename, ShapeFiles.eNew.shpOpen);
            }
            catch
            {
                nb_error++;
                return;
            }

            int nb_attributs = MyShape.ShapeFields.Count;
            int nb_points = MyShape.RecordCount;

            // On choisi arbitrairement le 1er polygonne du sous groupe de poly
            int index_part = 1;
            int refresh = 0;

            // Parcourt des Shapes
            for (int ind_shp = 1; ind_shp <= MyShape.RecordCount; ind_shp++)
            {
                refresh++;
                if (refresh == 1000)
                {
                    Application.DoEvents();
                    refresh = 0;
                }

                Plant plant = new Plant();

                MyShape.MoveTo(ind_shp);

                int n = MyShape.Parts[index_part].Begins;
                double Longitude = MyShape.Vertices[n].X_Cord;
                double Latitude = MyShape.Vertices[n].Y_Cord;

                string HeightRang = "";

                plant.position = new double[3] { Longitude, Latitude, 0 };

                //attributs
                foreach (ShapeField field in MyShape.ShapeFields)
                {
                    string stgField = field.FieldName;

                    if (stgField.ToUpper().IndexOf("Group".ToUpper()) != -1)
                    {
                        if (field.Value.ToString().IndexOf("1") != -1) plant.group = Group.broadleaf;
                        if (field.Value.ToString().IndexOf("2") != -1) plant.group = Group.palm;
                        if (field.Value.ToString().IndexOf("3") != -1) plant.group = Group.conifer;
                        if (field.Value.ToString().IndexOf("4") != -1) plant.group = Group.shrub;
                    }

                    if (stgField.ToUpper().IndexOf("Species".ToUpper()) != -1) if (field.Value != null) plant.species = field.Value.ToString(); else plant.species = "";

                    if (stgField.ToUpper().IndexOf("HeightRang".ToUpper()) != -1)
                    {
                        HeightRang = field.Value.ToString();
                    }
                }

                //Height Range fonction du group
                if (plant.group == Group.broadleaf || plant.group == Group.conifer)
                {
                    if (HeightRang.IndexOf("1") != -1) plant.height_range = new float[2] { 8, 20 };
                    if (HeightRang.IndexOf("2") != -1) plant.height_range = new float[2] { 20, 35 };
                    if (HeightRang.IndexOf("3") != -1) plant.height_range = new float[2] { 30, 45 };
                }
                if (plant.group == Group.palm)
                {
                    if (HeightRang.IndexOf("1") != -1) plant.height_range = new float[2] { 15, 25 };
                    if (HeightRang.IndexOf("2") != -1) plant.height_range = new float[2] { 30, 40 };
                    if (HeightRang.IndexOf("3") != -1) plant.height_range = new float[2] { 35, 50 };
                }
                if (plant.group == Group.shrub)
                {
                    plant.height_range = new float[2] { 2, 10 };
                }

                this.cultivation.List_plant.Add(plant);
            }
        }

        public string[] GetNextListlines(string[] lines, int index)
        {
            int start_list_line = index;
            int end_list_line = lines.Length - 1;

            for (int i = start_list_line+1; i < lines.Length; i++)
            {
                if (lines[i].IndexOf("<[list_", 0) != -1)
                {
                    end_list_line = i - 1;
                }
            }

            int lenght = end_list_line - start_list_line;
            string[] list_lines = new string[lenght];

            // add selected lines to list_lines
            int new_index = 0;
            for (int i = start_list_line; i < end_list_line; i++)
            {
                list_lines[new_index] = lines[i];
                new_index++;
            }

            return list_lines;
        }

        public void PopulateListFromlines(string[] lines)
        {
            int index = 1;
            string[] words;

            while (index < lines.Length)
            {
                words = decodeline(lines[index]);
                // Si c'est une definition de "Plant"
                if (words[0].IndexOf("plant") == 0)
                {
                    string[] plant_lines = GetSection(lines, index);
                    index += plant_lines.Length;
                    this.cultivation.List_plant.Add(DecodePlant(plant_lines));
                }

                // buildings
                if (words[0].IndexOf("building") == 0)
                {
                    string[] objet_lines = GetSection(lines, index);
                    index += objet_lines.Length;
                    if (objet_lines.Length < 1)
                    {
                        bool stop = true;
                    }
                    this.cultivation.List_building.Add(DecodeBuilding(objet_lines));
                }

                // lights
                if (words[0].IndexOf("light") == 0)
                {
                    string[] objet_lines = GetSection(lines, index);
                    index += objet_lines.Length;
                    this.cultivation.List_light.Add(DecodeLight(objet_lines));
                }

                index++;
            }
        }

        public string[] GetSection(string[] lines, int start_index)
        {
            List<string> section = new List<string>();

            for (int i = start_index; i < lines.Length; i++)
            {
                if (lines[i].Trim().IndexOf(">", 0) == 0 && lines[i].Trim().Length < 2)
                {
                    break;
                }
                section.Add(lines[i]);
            }

            return section.ToArray();
        }

        public Light DecodeLight(string[] lines)
        {
            string[] words = new string[3];

            Light objet = new Light(); ;

            foreach (string line in lines)
            {
                words = decodeline(line);

                //Position
                if (words[1].IndexOf("position", 0) != -1)
                {
                    objet.position = decodePosition(words[2]);
                }

                //Color
                if (words[1].IndexOf("color", 0) != -1)
                {
                    objet.color = decodevector3(words[2]);
                }

                //Intensity
                if (words[1].IndexOf("intensity", 0) != -1)
                {
                    objet.intensity = float.Parse(words[2]);
                }

                //flashing
                if (words[1].IndexOf("flashing", 0) != -1)
                {
                    objet.flashing = decodevector4(words[2]);
                }

                //Floor
                if (words[1].IndexOf("group_index", 0) != -1)
                {
                    objet.group_index = int.Parse(words[2]);
                }

            }

            return objet;
        }

        public Building DecodeBuilding(string[] lines)
        {
            string[] words = new string[3];

            Building objet = new Building();

            foreach (string line in lines)
            {
                words = decodeline(line);

                //Position
                if (words[1].IndexOf("position", 0) != -1)
                {
                    objet.position = decodePosition(words[2]);
                }

                //Direction
                if (words[1].IndexOf("direction", 0) != -1)
                {
                    objet.direction = float.Parse(words[2]);
                }

                //Length
                if (words[1].IndexOf("length", 0) != -1)
                {
                    objet.lenght = float.Parse(words[2]);
                }

                //Direction
                if (words[1].IndexOf("width", 0) != -1)
                {
                    objet.width = float.Parse(words[2]);
                }

                //Floor
                if (words[1].IndexOf("floors", 0) != -1)
                {
                    objet.floor = int.Parse(words[2]);
                }

                //Roof
                if (words[1].IndexOf("group", 0) != -1)
                {
                    objet.roof = (Roof)Enum.Parse(typeof(Roof), words[2]);
                }

                //Usage
                if (words[1].IndexOf("group", 0) != -1)
                {
                    objet.usage = (Usage)Enum.Parse(typeof(Usage), words[2]);
                }
            }

            if (objet.position == null)
            {
                bool stop = true;
            }
            //test
            //List<PointD> test = objet.GetCoord();

            return objet;
        }
        
        public Plant DecodePlant(string[]lines)
        {
            string[] words = new string[3];

            Plant plant = new Plant();

            foreach (string line in lines)
            {
                words = decodeline(line);
                
                //Position
                if (words[1].IndexOf("position",0) != -1)
                {
                    plant.position = decodePosition(words[2]);
                }

                //height range
                if (words[1].IndexOf("height_range", 0) != -1)
                {
                    plant.height_range = decodeRange(words[2]);
                }

                //group
                if (words[1].IndexOf("group", 0) != -1)
                {
                    plant.group = (Group)Enum.Parse(typeof(Group), words[2]);
                }

                //Species
                if (words[1].IndexOf("species", 0) != -1)
                {
                    if (words[2].Length > 0)
                    {
                        plant.species = words[2];
                    }
                }
            }

            return plant;
        }

        public void Read(string TOCpath)
        {
            string[] lines = System.IO.File.ReadAllLines(TOCpath);
            this.FullName = TOCpath;

            // File name
            string[] words = decodeline(lines[0]);
            this.Name = words[2];

            // coordinate system
            words = decodeline(lines[2]);
            cultivation.coordinate_system = (Coordinate_system)Enum.Parse(typeof(Coordinate_system), words[2]);

            int refresh = 0;

            //**** start decode lists ****
            int start_list_line = 3;

            while (start_list_line < lines.Length)
            {
                string[] listlines = GetNextListlines(lines, start_list_line);
                // do work on listlines
                if (listlines.Length > 0)
                {
                    PopulateListFromlines(listlines);
                    start_list_line += listlines.Length;
                }
                start_list_line++;
            }
        }
       
       public Cultivation[] GetCultivationGrid(Cultivation cultivation, int level)
        {
            AFS2_Grid grid = new AFS2_Grid(level);
            grid.FillCells_INBBox(cultivation.Xmin, cultivation.Xmax, cultivation.Ymin, cultivation.Ymax, AFS2_Grid.SelectMode.Total);

            Cultivation[] cultivationGrid = new Cultivation[grid.Cells.Count];
            for (int i = 0; i < cultivationGrid.Length; i++)
            {
                cultivationGrid[i] = new Cultivation();
                cultivationGrid[i].Name = grid.Cells[i].AFS2Name;
            }

            // plant
            foreach (Plant plant in cultivation.List_plant)
            {
                for (int i = 0; i < grid.Cells.Count; i++)
                {
                    PointD coord = new PointD(plant.position[0], plant.position[1]);
                    if (grid.Cells[i].ContainPoint(coord))
                    {
                        cultivationGrid[i].List_plant.Add(plant);
                        break;
                    }
                }
            }

            // building
            foreach (Building build in cultivation.List_building)
            {
                for (int i = 0; i < grid.Cells.Count; i++)
                {
                    PointD coord = new PointD(build.position[0], build.position[1]);
                    if (grid.Cells[i].ContainPoint(coord))
                    {
                        cultivationGrid[i].List_building.Add(build);
                        break;
                    }
                }
            }

            // light
            foreach (Light light in cultivation.List_light)
            {
                for (int i = 0; i < grid.Cells.Count; i++)
                {
                    PointD coord = new PointD(light.position[0], light.position[1]);
                    if (grid.Cells[i].ContainPoint(coord))
                    {
                        cultivationGrid[i].List_light.Add(light);
                        break;
                    }
                }
            }

            return cultivationGrid;
        }        

        public void Read_old(string TOCpath)
        {
            string[] lines = System.IO.File.ReadAllLines(TOCpath);
            this.FullName = TOCpath;

            // File name
            string[] words = decodeline(lines[0]);
            this.Name = words[2];

            // coordinate system
            words = decodeline(lines[2]);
            cultivation.coordinate_system = (Coordinate_system)Enum.Parse(typeof(Coordinate_system), words[2]);

            int refresh = 0;

            //**** start decode lists ****
            int start_list_line = 3;

            while (start_list_line < lines.Length)
            {
                string[] listlines = GetNextListlines(lines, start_list_line);
                // do work on listlines
                // ...
                start_list_line += listlines.Length;
            }


            for (int i = 3; i < lines.Length; i++)
            {
                refresh++;
                if (refresh == 1000)
                {
                    Application.DoEvents();
                    refresh = 0;
                }
                words = decodeline(lines[i]);

                // Si c'est une definition de "Plant"
                if (words[0].IndexOf("plant") == 0)
                {
                    Plant plant = new Plant();

                    //position
                    words = decodeline(lines[i + 1]);
                    plant.position = decodePosition(words[2]);

                    //height range
                    words = decodeline(lines[i + 2]);
                    plant.height_range = decodeRange(words[2]);

                    //group
                    words = decodeline(lines[i + 3]);
                    plant.group = (Group)Enum.Parse(typeof(Group), words[2]);

                    //species
                    words = decodeline(lines[i + 4]);
                    if (words[2].Length > 0)
                    {
                        plant.species = words[2];
                    }

                    // add info
                    this.Info.plantInfo.AddInfo(plant);

                    this.cultivation.List_plant.Add(plant);

                    i = i + 5;
                }
            }

        }

        public void LimitToc(int MaxpercentForest, int MaxpercentBosquet, int MaxpercentSmall)
        {
            List<Plant> new_list = new List<Plant>();

            Random rnd = new Random();

            foreach (Plant plant in this.cultivation.List_plant)
            {
                int percent = rnd.Next(0, 100);
                bool keep = true;
                if (plant.height_range[1] > 40)
                {
                    if (percent > MaxpercentForest) keep = false;
                }
                if (plant.height_range[1] > 25 && plant.height_range[1] <= 40)
                {
                    if (percent > MaxpercentBosquet) keep = false;
                }
                if (plant.height_range[1] <= 25)
                {
                    if (percent > MaxpercentSmall) keep = false;
                }

                if (keep) new_list.Add(plant);
                Application.DoEvents();
            }
            this.cultivation.List_plant = new_list;
        }

        public void Write(string TOCpath)
        {
            StringBuilder stg = new StringBuilder();

            stg.AppendLine("<[file]" + "[]" + "[" + Name + "]");
            stg.AppendLine("  " + "<[cultivation]" + "[]" + "[]");
            stg.AppendLine("    " + "<[string8]" + "[coordinate_system]" + "[" + cultivation.coordinate_system + "]>");

            if (this.cultivation.List_plant.Count > 0)
            {
                stg.AppendLine("    " + "<[list_plant]" + "[plant_list]" + "[]");

                foreach (Plant plant in cultivation.List_plant)
                {
                    stg.AppendLine("      " + "<[plant]" + "[element]" + "[0]");

                    stg.AppendLine("        " + "<[vector3_float64]" + "[position]" + "[" + String.Format("{0:0.000000}", plant.position[0]) + " " + String.Format("{0:0.000000}", plant.position[1]) + " " + plant.position[2] + "]>");
                    stg.AppendLine("        " + "<[vector2_float32]" + "[height_range]" + "[" + String.Format("{0:0.000000}", plant.height_range[0]) + " " + String.Format("{0:0.000000}", plant.height_range[1]) + "]>");
                    stg.AppendLine("        " + "<[string8]" + "[group]" + "[" + plant.group + "]>");
                    stg.AppendLine("        " + "<[string8]" + "[species]" + "[" + plant.species + "]>");

                    stg.AppendLine("      " + ">");
                }

                stg.AppendLine("    " + ">");
            }

            stg.AppendLine("  " + ">");

            stg.AppendLine(">");

            // write file
            System.IO.File.WriteAllText(TOCpath, stg.ToString());
        }

        public void Write()
        {
            StringBuilder stg = new StringBuilder();

            stg.AppendLine("<[file]" + "[]" + "[" + Name + "]");
            stg.AppendLine("  " + "<[cultivation]" + "[]" + "[]");
            stg.AppendLine("    " + "<[string8]" + "[coordinate_system]" + "[" + cultivation.coordinate_system + "]>");

            // Plant
            if (this.cultivation.List_plant.Count > 0)
            {
                stg.AppendLine("    " + "<[list_plant]" + "[plant_list]" + "[]");

                foreach (Plant plant in cultivation.List_plant)
                {
                    stg.AppendLine("      " + "<[plant]" + "[element]" + "[0]");

                    stg.AppendLine("        " + "<[vector3_float64]" + "[position]" + "[" + String.Format("{0:0.000000}", plant.position[0]) + " " + String.Format("{0:0.000000}", plant.position[1]) + " " + plant.position[2] + "]>");
                    stg.AppendLine("        " + "<[vector2_float32]" + "[height_range]" + "[" + String.Format("{0:0.000000}", plant.height_range[0]) + " " + String.Format("{0:0.000000}", plant.height_range[1]) + "]>");
                    stg.AppendLine("        " + "<[string8]" + "[group]" + "[" + plant.group + "]>");
                    stg.AppendLine("        " + "<[string8]" + "[species]" + "[" + plant.species + "]>");

                    stg.AppendLine("      " + ">");
                }

                stg.AppendLine("    " + ">");
            }

            //Building
            if (this.cultivation.List_building.Count > 0)
            {
                stg.AppendLine("    " + "<[list_building]" + "[building_list]" + "[]");

                foreach (Building objet in cultivation.List_building)
                {
                    stg.AppendLine("      " + "<[building]" + "[element]" + "[0]");

                    stg.AppendLine("        " + "<[vector3_float64]" + "[position]" + "[" + String.Format("{0:0.000000}", objet.position[0]) + " " + String.Format("{0:0.000000}", objet.position[1]) + " " + String.Format("{0:0.000000}", objet.position[2]) + "]>");
                    stg.AppendLine("        " + "<[float32]" + "[direction]" + "[" + String.Format("{0:0.000000}", objet.direction) + "]>");
                    stg.AppendLine("        " + "<[float32]" + "[length]" + "[" + String.Format("{0:0.000000}", objet.lenght) + "]>");
                    stg.AppendLine("        " + "<[float32]" + "[width]" + "[" + String.Format("{0:0.000000}", objet.width) + "]>");
                    stg.AppendLine("        " + "<[int32]" + "[floors]" + "[" + objet.floor + "]>");
                    stg.AppendLine("        " + "<[string8]" + "[roof]" + "[" + objet.roof + "]>");
                    stg.AppendLine("        " + "<[string8]" + "[usage]" + "[" + objet.usage + "]>");

                    stg.AppendLine("      " + ">");
                }

                stg.AppendLine("    " + ">");
            }

            //Light
            if (this.cultivation.List_light.Count > 0)
            {
                stg.AppendLine("    " + "<[list_light]" + "[light_list]" + "[]");

                foreach (Light objet in cultivation.List_light)
                {
                    stg.AppendLine("      " + "<[light]" + "[element]" + "[0]");

                    stg.AppendLine("        " + "<[vector3_float64]" + "[position]" + "[" + String.Format("{0:0.000000}", objet.position[0]) + " " + String.Format("{0:0.000000}", objet.position[1]) + " " + String.Format("{0:0.000000}", objet.position[2]) + "]>");
                    stg.AppendLine("        " + "<[vector3_float32]" + "[color]" + "[" + String.Format("{0:0.000000}", objet.color[0]) + " " + String.Format("{0:0.000000}", objet.color[1]) + " " + String.Format("{0:0.000000}", objet.color[2]) + "]>");
                    stg.AppendLine("        " + "<[float32]" + "[intensity]" + "[" + String.Format("{0:0.000000}", objet.intensity) + "]>");
                    stg.AppendLine("        " + "<[vector4_float32]" + "[flashing]" + "[" + String.Format("{0:0.000000}", objet.flashing[0]) + " " + String.Format("{0:0.000000}", objet.flashing[1]) + String.Format("{0:0.000000}", objet.flashing[2]) + " " + String.Format("{0:0.000000}", objet.flashing[3]) + "]>");
                    stg.AppendLine("        " + "<[uint32]" + "[group_index]" + "[" + objet.group_index + "]>");

                    stg.AppendLine("      " + ">");
                }

                stg.AppendLine("    " + ">");
            }


            stg.AppendLine("  " + ">");

            stg.AppendLine(">");

            // write file
            System.IO.File.WriteAllText(this.FullName, stg.ToString());
        }

        public void ReplaceGroup(Group SourceGrp, Group ReplaceGrp)
        {
            foreach (Plant plant in this.cultivation.List_plant)
            {
                if (plant.group == SourceGrp) plant.group = ReplaceGrp;
            }
        }

        public string[] decodeline(string line)
        {
            string[] words = new string[3];

            int start = 0;
            int end = 1;

            for (int i = 0; i < 3; i++)
            {
                start = line.IndexOf("[", start) + 1;
                end = line.IndexOf("]", start);

                if (start != -1 && end != -1)
                {
                    words[i] = line.Substring(start, end - start);
                }
                else return new string[3] { "", "", "" };
            }

            return words;
        }

        public double[] decodePosition(string word)
        {
            int space = word.IndexOf(" ");
            int space2 = word.IndexOf(" ", space + 1);
            string word1 = word.Substring(0, space);
            string word2 = word.Substring(space + 1, space2 - space - 1);
            string word3 = word.Substring(space2 + 1, word.Length - space2 - 1);

            return new double[3] { double.Parse(word1), double.Parse(word2), double.Parse(word3) };
        }

        public float[] decodevector3(string word)
        {
            int space = word.IndexOf(" ");
            int space2 = word.IndexOf(" ", space + 1);

            string word1 = word.Substring(0, space);
            string word2 = word.Substring(space + 1, space2 - space - 1);
            string word3 = word.Substring(space2 + 1, word.Length - space2 - 1);

            return new float[3] { float.Parse(word1), float.Parse(word2), float.Parse(word3) };
        }

        public float[] decodevector4(string word)
        {
            int space = word.IndexOf(" ");
            int space2 = word.IndexOf(" ", space + 1);
            int space3 = word.IndexOf(" ", space2 + 1);

            string word1 = word.Substring(0, space);
            string word2 = word.Substring(space + 1, space2 - space - 1);
            string word3 = word.Substring(space2 + 1, space3 - space2 - 1);
            string word4 = word.Substring(space3 + 1, word.Length - space3 - 1);

            return new float[4] { float.Parse(word1), float.Parse(word2), float.Parse(word3), float.Parse(word4) };
        }

        public double[] decodeFloat(string word)
        {
            int space = word.IndexOf(" ");
            int space2 = word.IndexOf(" ", space + 1);
            string word1 = word.Substring(0, space);
            string word2 = word.Substring(space + 1, space2 - space - 1);
            string word3 = word.Substring(space2 + 1, word.Length - space2 - 1);

            return new double[3] { double.Parse(word1), double.Parse(word2), double.Parse(word3) };
        }

        public float[] decodeRange(string word)
        {
            int space = word.IndexOf(" ");
            string word1 = word.Substring(0, space);
            string word2 = word.Substring(space + 1, word.Length - space - 1);

            return new float[2] { float.Parse(word1), float.Parse(word2) };
        }

        // Sort Randomly
        public void SortRandomly()
        {
            var rnd = new Random();
            this.cultivation.List_plant = this.cultivation.List_plant.OrderBy(x => rnd.Next()).ToList();
            Application.DoEvents();
        }

    }

    public class Cultivation
    {
        public Coordinate_system coordinate_system { get; set; }
        public List<Plant> List_plant { get; set; }
        public List<Building> List_building { get; set; }
        public List<Light> List_light { get; set; }

        public string Name { get; set; }

        public double Xmin { get; set; }
        public double Xmax { get; set; }
        public double Ymin { get; set; }
        public double Ymax { get; set; }

        // *****************
        public Cultivation()
        {
            coordinate_system = Coordinate_system.lonlat;
            List_plant = new List<Plant>();
            List_building = new List<Building>();
            List_light = new List<Light>();
        }

        public void SetBoundingBox()
        {
            bool first = true;
            int error = 0;

            foreach (Plant plant in List_plant)
            {
                if (first || plant.position[0] < Xmin) Xmin = plant.position[0];
                if (first || plant.position[1] < Ymin) Ymin = plant.position[1];
                if (first || plant.position[0] > Xmax) Xmax = plant.position[0];
                if (first || plant.position[1] > Ymax) Ymax = plant.position[1];
                first = false;
            }

            first = true;

            foreach (Building objet in List_building)
            {
                if (objet.position != null)
                {
                    if (first || objet.position[0] < Xmin) Xmin = objet.position[0];
                    if (first || objet.position[1] < Ymin) Ymin = objet.position[1];
                    if (first || objet.position[0] > Xmax) Xmax = objet.position[0];
                    if (first || objet.position[1] > Ymax) Ymax = objet.position[1];
                    first = false;
                }
                else error++;
            }

            first = true;

            foreach (Light objet in List_light)
            {
                if (first || objet.position[0] < Xmin) Xmin = objet.position[0];
                if (first || objet.position[1] < Ymin) Ymin = objet.position[1];
                if (first || objet.position[0] > Xmax) Xmax = objet.position[0];
                if (first || objet.position[1] > Ymax) Ymax = objet.position[1];
                first = false;
            }
        }

    }

    //******************* ENUM *************
    public enum Coordinate_system
    {
        lonlat = 0
    }

    public enum Group
    {
        broadleaf = 0,
        conifer = 1,
        palm = 2,
        shrub = 3
    }

    public enum Species
    {
        none = 0
    }

    public enum Roof
    {
        flat = 0,
        gable = 1
    }

    public enum Usage
    {
        residential = 0,
        industrial = 1
    }
    //**********************

    public class Plant
    {
        public double[] position { get; set; }
        public float[] height_range { get; set; }
        public Group group { get; set; }
        public string species { get; set; }

        public Plant(PointD coord, float minHeight, float maxHeight, Group group, string species)
        {
            this.position = new double[3] { coord.X, coord.Y, coord.Z };
            this.height_range = new float[2] { minHeight, maxHeight };
            this.group = group;
            this.species = species;
        }

        public Plant()
        {

        }

    }

    public class Building
    {
        public double[] position { get; set; }
        public float direction { get; set; }
        public float lenght { get; set; }
        public float width { get; set; }
        public int floor { get; set; }
        public Roof roof { get; set; }
        public Usage usage { get; set; }

        public Building(PointD coord, float Direction, float Lenght, float Width, int floor, Roof roofType, Usage usageType)
        {
            this.position = new double[3] { coord.X, coord.Y, coord.Z };
            this.direction = Direction;
            this.lenght = Lenght;
            this.width = Width;
            this.roof = roofType;
            this.usage = usageType;
        }

        public Building()
        { }

        public List<PointD> GetCoord()
        {
            // position = centre du rectangle !!!

            PointD Center = new PointD(position[0], position[1]);

            // calcul vecteur
            Vecteur3D vect = new Vecteur3D(direction);
            Vecteur3D vect_ortho = vect.GetVecteurOrtho_AxeZ();

            PointD pointInter1 = Tools.GetPointB_xmeter_from_PointA(Center, this.lenght/2, vect);
            PointD pointInter2 = Tools.GetPointB_xmeter_from_PointA(Center, -this.lenght / 2, vect);

            PointD A = Tools.GetPointB_xmeter_from_PointA(pointInter1, this.width/2, vect_ortho);
            PointD B = Tools.GetPointB_xmeter_from_PointA(pointInter1, -this.width/2, vect_ortho);

            PointD C = Tools.GetPointB_xmeter_from_PointA(pointInter2, -this.width/2, vect_ortho);
            PointD D = Tools.GetPointB_xmeter_from_PointA(pointInter2, this.width/2, vect_ortho);
            
            // string debug = A.X + "," + A.Y + " " + B.X + "," + B.Y + " " + C.X + "," + C.Y + " " + D.X + "," + D.Y + " " + A.X + "," + A.Y + " ";

            return new List<PointD>() { A, B, C, D };
        }

    }

    public class Light
    {
        public double[] position { get; set; }
        public float[] color { get; set; }
        public float intensity { get; set; }
        public float[] flashing { get; set; }
        public int group_index { get; set; }

        public Light(PointD coord, float[] Color, float Intensity, float[] Flashing, int Group_Index)
        {
            this.position = new double[3] { coord.X, coord.Y, coord.Z };
            this.color = Color;
            this.intensity = Intensity;
            this.flashing = Flashing;
            this.group_index = Group_Index;
        }

        public Light()
        { }
    }

    public class TOCFileInfo
    {
        public PlantInfo plantInfo { get; set; }

        public TOCFileInfo()
        {
            plantInfo = new PlantInfo();
        }
    }

    
    public class GroupInfo
    {
        public Group group { get; set; }
        public List<HeightRangeInfo> heightRangeinfos { get; set; }
        public int Count { get; set; }

        public GroupInfo()
        {
            group = new Group();
            heightRangeinfos = new List<HeightRangeInfo>();
        }
    }

    public class HeightRangeInfo
    {
        public float[] height_range { get; set; }
        public int Count { get; set; }

        public HeightRangeInfo()
        {
            
        }
    }

    public class PlantInfo
    {
        public List<GroupInfo> groupinfos { get; set; }
        public int Count { get; set; }

        public PlantInfo()
        {
            groupinfos = new List<GroupInfo>();
            
        }

        public void AddInfo(Plant plant)
        {
            // Gestion de la liste des categories d'arbre
            bool GroupExist = false;
            foreach (GroupInfo groupinfo in groupinfos)
            {
                if (groupinfo.group == plant.group)
                {
                    groupinfo.Count++;
                    GroupExist = true;

                    bool HeightRangeExist = false;
                    foreach (HeightRangeInfo height in groupinfo.heightRangeinfos)
                    {
                        if (height.height_range[0] == plant.height_range[0] && height.height_range[1] == plant.height_range[1])
                        {
                            height.Count++;
                            HeightRangeExist = true;
                            break;
                        }
                    }

                    if (!HeightRangeExist)
                    {
                        // Gestion de la list des HeightRange
                        HeightRangeInfo height = new HeightRangeInfo();
                        height.height_range = plant.height_range;
                        height.Count++;
                        groupinfo.heightRangeinfos.Add(height);
                    }

                    break;
                } 
            }

            if (!GroupExist)
            {
                GroupInfo info = new GroupInfo();
                info.group = plant.group;
                info.Count++;

                // Gestion de la list des HeightRange
                HeightRangeInfo height = new HeightRangeInfo();
                height.height_range = plant.height_range;
                height.Count++;
                info.heightRangeinfos.Add(height);

                groupinfos.Add(info);               
            }

            this.Count++;
        }
    }

    // ************************* TMC Files **************************

    public class TMCFile
    {
        public bool write_images_with_mask { get; set; }
        public bool write_raw_files { get; set; }
        public bool write_ttc_files { get; set; }
        public bool always_overwrite { get; set; }
        public bool do_heightmaps { get; set; }
        public string folder_destination_ttc { get; set; }
        public string folder_destination_raw { get; set; }
        public string folder_source_files { get; set; }

        List<TMColormap_Regions> Regions { get; set; }

        public TMCFile()
        {
            Regions = new List<TMColormap_Regions>();
        }

        public void write()
        {
            StringBuilder stg = new StringBuilder();

            stg.AppendLine("<[file]" + "[]" + "[" + "]");
            stg.AppendLine("/t" + "<[tmcolormap_regions]" + "[]" + "[]");
                stg.AppendLine("/t/t" + "<[string8]" + "[folder_source_files]" + "[" + folder_source_files + "]>");
                stg.AppendLine("/t/t" + "<[string8]" + "[folder_destination_ttc]" + "[" + folder_destination_ttc + "]>");
                stg.AppendLine("/t/t" + "<[string8]" + "[folder_source_files]" + "[" + folder_destination_raw + "]>");
                stg.AppendLine("/t/t" + "<[bool]" + "[write_ttc_files]" + "[" + write_ttc_files.ToString() + "]>");
                stg.AppendLine("/t/t" + "<[bool]" + "[write_raw_files]" + "[" + write_raw_files.ToString() + "]>");
                stg.AppendLine("/t/t" + "<[bool]" + "[do_heightmaps]" + "[" + do_heightmaps + "]>");
                stg.AppendLine("/t/t" + "<[bool]" + "[always_overwrite]" + "[" + always_overwrite + "]>");
                stg.AppendLine("/t/t" + "<[bool]" + "[write_images_with_mask]" + "[" + write_images_with_mask + "]>");
                stg.AppendLine("/t/t" + "<[bool]" + "[always_overwrite]" + "[" + always_overwrite + "]>");
            stg.AppendLine("");
                stg.AppendLine("/t/t" + "<[list]" + "[region_list]" + "[]");

            
        }

    }

    public class TMColormap_Regions
    {
        public int Level { get; set; }
        public double[] lonlat_min { get; set; }
        public double[] lonlat_max { get; set; }
        public bool write_images_with_mask { get; set; }

        public TMColormap_Regions()
        {

        }

    }


    // **************** Autres

    public class Vecteur3D
    {
        public Vecteur3D()
        {
            this.X = 0;
            this.Y = 0;
            this.Z = 0;
        }

        public Vecteur3D(Double x, Double y, Double z)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
        }

        public Vecteur3D(Vecteur3D pointA, Vecteur3D pointB)
        {
            this.X = pointB.X - pointA.X;
            this.Y = pointB.Y - pointA.Y;
            this.Z = pointB.Z - pointA.Z;
        }

        public Vecteur3D(Point3D pointA, Point3D pointB)
        {
            this.X = pointB.X - pointA.X;
            this.Y = pointB.Y - pointA.Y;
            this.Z = pointB.Z - pointA.Z;
        }

        public Vecteur3D(double heading)
        {
            double radians = heading * Math.PI / 180.0;

            this.X = Math.Cos(radians);
            this.Y = Math.Sin(radians);
            this.Z = 0;
        }

        public double GetHeading()
        {
            double angle = Math.Atan2(this.Y, this.X) * 180.0 / Math.PI;
            return angle;
        }

        public Double X { get; set; }
        public Double Y { get; set; }
        public Double Z { get; set; }

        public bool Egal(Vecteur3D vector)
        {
            if ((this.X == vector.X) && (this.Y == vector.Y) && (this.Z == vector.Z)) return true;
            return false;
            // renvoi true si identique
        }

        public void turn90_AxeZ()
        {
            this.X = this.Y;
            this.Y = -this.X;
        }

        public Vecteur3D GetVecteurOrtho_AxeZ()
        {
            Vecteur3D ortho = new Vecteur3D(this.Y, -this.X,0);
            return ortho;
        }

        public void turn180_AxeZ()
        {
            this.X = -this.X;
            this.Y = -this.Y;
        }

        public static Vecteur3D turn90_AxeZ(Vecteur3D v1)
        {
            return new Vecteur3D(v1.Y, -v1.X, v1.Z);
        }

        public static Vecteur3D turn180_AxeZ(Vecteur3D v1)
        {
            return new Vecteur3D(-v1.X, -v1.Z, v1.Z);
        }


        public void normalize()
        {
            double length = Math.Sqrt((this.X * this.X) + (this.Y * this.Y) + (this.Z * this.Z));
            this.X /= length;
            this.Y /= length;
            this.Z /= length;
        }


        public void simplify()
        {
            if ((this.X == 0 || Math.Abs(this.X / 2) >= 1) && (this.Y == 0 || Math.Abs(this.Y / 2) >= 1) && (this.Z == 0 || Math.Abs(this.Z / 2) >= 1))
            {
                if (this.X != 0) this.X = this.X / 2;
                if (this.Y != 0) this.Y = this.Y / 2;
                if (this.Z != 0) this.Z = this.Z / 2;
            }
        } // Obsolete avant normalize

        public static Vecteur3D Getnormal(Vecteur3D pointA, Vecteur3D pointB, Vecteur3D pointC)
        {
            // calcul du vecteur de la normale d'une face 3D
            // On prend arbitrairement deux vecteurs en fonction de 3 points de la face pointA, pointB, pointC

            // Creation de deux vecteur par le produit scalaire D de A et B et E entre A et C
            Vecteur3D D = new Vecteur3D((pointB.X - pointA.X), (pointB.Y - pointA.Y), (pointB.Z - pointA.Z));
            Vecteur3D E = new Vecteur3D((pointC.X - pointA.X), (pointC.Y - pointA.Y), (pointC.Z - pointA.Z));

            // coordonnées du vecteur normale
            Vecteur3D Normal = new Vecteur3D();
            Normal.X = (pointB.Y - pointA.Y) * (pointC.Z - pointA.Z) - (pointB.Z - pointA.Z) * (pointC.Y - pointA.Y);
            Normal.Y = (pointB.Z - pointA.Z) * (pointC.X - pointA.X) - (pointB.X - pointA.X) * (pointC.Z - pointA.Z);
            Normal.Z = (pointB.X - pointA.X) * (pointC.Y - pointA.Y) - (pointB.Y - pointA.Y) * (pointC.X - pointA.X);

            return Normal;
        }

        public static Vecteur3D CrossProduct(Vecteur3D v1, Vecteur3D v2)
        {
            double x, y, z;

            x = v1.Y * v2.Z - v2.Y * v1.Z;
            y = (v1.X * v2.Z - v2.X * v1.Z) * -1;
            z = v1.X * v2.Y - v2.X * v1.Y;

            Vecteur3D rtnvector = new Vecteur3D(x, y, z);
            return rtnvector;
        }

        public static double DotProduct(Vecteur3D v1, Vecteur3D v2)
        {
            return
            (
               v1.X * v2.X +
               v1.Y * v2.Y +
               v1.Z * v2.Z
            );
        }

        public static Vecteur3D Getnormal(Vertice pointA, Vertice pointB, Vertice pointC)
        {
            // calcul du vecteur de la normale d'une face 3D
            // On prend arbitrairement deux vecteurs en fonction de 3 points de la face pointA, pointB, pointC

            // Creation de deux vecteur par le produit scalaire D de A et B et E entre A et C
            Vecteur3D D = new Vecteur3D((pointB.X - pointA.X), (pointB.Y - pointA.Y), (pointB.Z - pointA.Z));
            Vecteur3D E = new Vecteur3D((pointC.X - pointA.X), (pointC.Y - pointA.Y), (pointC.Z - pointA.Z));

            // coordonnées du vecteur normale
            Vecteur3D Normal = new Vecteur3D();
            Normal.X = (pointB.Y - pointA.Y) * (pointC.Z - pointA.Z) - (pointB.Z - pointA.Z) * (pointC.Y - pointA.Y);
            Normal.Y = (pointB.Z - pointA.Z) * (pointC.X - pointA.X) - (pointB.X - pointA.X) * (pointC.Z - pointA.Z);
            Normal.Z = (pointB.X - pointA.X) * (pointC.Y - pointA.Y) - (pointB.Y - pointA.Y) * (pointC.X - pointA.X);

            return Normal;
        }

        public static Vecteur3D GetVector2Origin(Vecteur3D FirstNode)
        {
            // Calcul de la translation à appliquer à tous les Nodes de la base pour translater le premier Node sur l'origine du repère local (0,0,0)

            Vecteur3D translation = new Vecteur3D();
            translation.X = -FirstNode.X;
            translation.Y = -FirstNode.Y;
            translation.Z = -FirstNode.Z;

            return translation;
        }

        public static Vecteur3D Translate2Origin(Vecteur3D Node, Vecteur3D toOriginVector)
        {
            // Renvoi les nouvelles coordonnées du Node après translation
            Vecteur3D retour = new Vecteur3D();
            retour.X = Node.X - toOriginVector.X;
            retour.Y = Node.Y - toOriginVector.Y;
            retour.Z = Node.Z - toOriginVector.Z;

            return retour;
        }


        public static Vecteur3D operator +(Vecteur3D A, Vecteur3D B)
        {
            return new Vecteur3D(A.X + B.X, A.Y + B.Y, A.Z + B.Z);
        }

        public static Vecteur3D operator -(Vecteur3D A, Vecteur3D B)
        {
            return new Vecteur3D(A.X - B.X, A.Y - B.Y, A.Z - B.Z);
        }

        public static bool operator ==(Vecteur3D A, Vecteur3D B)
        {
            return (A.X == B.X && A.Y == B.Y && A.Z == B.Z);
        }

        public static bool operator !=(Vecteur3D A, Vecteur3D B)
        {
            return !(A.X == B.X && A.Y == B.Y && A.Z == B.Z);
        }

    }

    public class Point3D
    {
        public Double X { get; set; }
        public Double Y { get; set; }
        public Double Z { get; set; }

        public Point3D()
        {
            this.X = 0;
            this.Y = 0;
            this.Z = 0;
        }

        public Point3D(Double x, Double y, Double z)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
        }

        public static Point3D MidPoint(Point3D A, Point3D B)
        {
            return new Point3D((A.X + B.X) / 2, (A.Y + B.Y) / 2, (A.Z + B.Z) / 2);
        }

        public static Point3D operator +(Point3D A, Point3D B)
        {
            return new Point3D(A.X + B.X, A.Y + B.Y, A.Z + B.Z);
        }

        public static Point3D operator -(Point3D A, Point3D B)
        {
            return new Point3D(A.X - B.X, A.Y - B.Y, A.Z - B.Z);
        }

        public static bool operator ==(Point3D A, Point3D B)
        {
            return (A.X == B.X && A.Y == B.Y && A.Z == B.Z);
        }

        public static bool operator !=(Point3D A, Point3D B)
        {
            return !(A.X == B.X && A.Y == B.Y && A.Z == B.Z);
        }

        public bool Egal(Point3D point3d)
        {
            if ((this.X == point3d.X) && (this.Y == point3d.Y) && (this.Z == point3d.Z)) return true;
            return false;
            // renvoi true si identique
        }

    }

    public class Vertice : IDisposable

    {
        public Vertice()
        {
            this.X = 0;
            this.Y = 0;
            this.Z = 0;

            this.Index_Face = new List<int>();
            this.Index_Normale = new List<int>();
            this.Index_VTexture = new List<int>();
        }

        public void Dispose()
        {
            this.Index_Face.Clear();
            this.Index_Normale.Clear();
            this.Index_VTexture.Clear();
        }

        public Vertice(Double x, Double y, Double z)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;

            this.Index_Face = new List<int>();
            this.Index_Normale = new List<int>();
            this.Index_VTexture = new List<int>();
        }

        public Double X { get; set; }
        public Double Y { get; set; }
        public Double Z { get; set; }

        public int Index { get; set; }

        public List<int> Index_Face { get; set; }
        public List<int> Index_Normale { get; set; }
        public List<int> Index_VTexture { get; set; }
    }

    // *************************************************
    // **** Areofly GRID

    class AFS2_Grid
    {

        public int Level { get; set; }
        private int nb_cellinEarthWidth { get; set; }


        public double Cell_Width { get; set; }
        public double Cell_Height { get; set; }
        public double Grid_Width { get; set; }

        public PointD[] Ref_points { get; set; }

        public List<Cellule> Cells { get; set; }

        public void Save2kml_Total()
        {
            // creation des polygones en fonction de la liste des coordonnées dans point
            List<SharpKml.Dom.Polygon> Polygones = new List<SharpKml.Dom.Polygon>();

            for (int line = 0; line < nb_cellinEarthWidth; line++)
            {
                for (int col = 0; col < nb_cellinEarthWidth; col++)
                {
                    PointD point1 = new PointD(this.Ref_points[col].X, this.Ref_points[line].Y);
                    PointD point2 = new PointD(this.Ref_points[col + 1].X, this.Ref_points[line + 1].Y);
                    Polygones.Add(GetPolygon(point1, point2));
                }
            }


            // creation du fichier kml
            SharpKml.Dom.Style stl = new SharpKml.Dom.Style();
            stl = CreateStyle(Color.FromArgb(30, 239, 164, 11), Color.FromArgb(255, 239, 164, 11), 3, 90, 1, 1);

            SharpKml.Dom.Document doc = new SharpKml.Dom.Document();
            doc.Name = "Aerofly SF2 GRID Level " + Level;
            doc.AddStyle(stl);

            foreach (SharpKml.Dom.Polygon polygon in Polygones)
            {
                // Creation du conteneur du polygone
                SharpKml.Dom.Placemark placemk = Polygon2placmk(polygon, polygon.Id, stl);

                StringBuilder stg = new StringBuilder();
                stg.AppendLine("Name: " + polygon.Id);
                stg.AppendLine("Ref Point: Latitude" + polygon.OuterBoundary.LinearRing.Coordinates.ElementAt(0).Latitude + " Longitude:" + polygon.OuterBoundary.LinearRing.Coordinates.ElementAt(0).Longitude);

                placemk.Description = new SharpKml.Dom.Description
                {
                    Text = stg.ToString()
                };

                doc.AddFeature(placemk);
            }


            SharpKml.Engine.KmlFile kml = SharpKml.Engine.KmlFile.Create(doc, true);

            string kml_filepath = @"C:\Users\Vogel10\source\repos\Raster_Converter\Raster_Converter\bin\Debug" + "\\" + "kmlgrid_level_" + Level + ".kml";

            using (var stream = System.IO.File.OpenWrite(kml_filepath))
            {
                kml.Save(stream);
            }
        }

        public void Save2kml_SelectedCells(string kml_filepath)
        {
            // creation du fichier kml
            SharpKml.Dom.Style stl = new SharpKml.Dom.Style();
            stl = CreateStyle(Color.FromArgb(30, 239, 164, 11), Color.FromArgb(255, 239, 164, 11), 3, 90, 1, 1);

            SharpKml.Dom.Document doc = new SharpKml.Dom.Document();
            doc.Name = "Aerofly SF2 GRID Level " + Level;
            doc.AddStyle(stl);

            //Creation des polygones au format kml
            foreach (Cellule c in this.Cells)
            {
                SharpKml.Dom.Polygon polygon = Cell2Polygon(c);

                // Creation du conteneur du polygone
                SharpKml.Dom.Placemark placemk = Polygon2placmk(polygon, polygon.Id, stl);

                StringBuilder stg = new StringBuilder();
                stg.AppendLine("<B><span style =\"color: RGB(0, 100, 255); \">" + "Name: " + "</span></B>" + c.AFS2Name);
                stg.AppendLine("<br><br>");
                stg.AppendLine("<B><span style =\"color: RGB(0, 100, 255); \">" + "Ref Point:" + "</span></B>");
                stg.AppendLine("<br>");
                stg.AppendLine("<B><span style =\"color: RGB(230, 20, 0); \">" + "Latitude: " + "</span></B>" + polygon.OuterBoundary.LinearRing.Coordinates.ElementAt(0).Latitude);
                stg.AppendLine("<br>");
                stg.AppendLine("<B><span style =\"color: RGB(230, 20, 0); \">" + "Longitude: " + "</span></B>" + polygon.OuterBoundary.LinearRing.Coordinates.ElementAt(0).Longitude);
                stg.AppendLine("<br><br>");
                stg.AppendLine("<span style =\"color: RGB(0, 100, 255); \">" + "<B>BoundingBox: </B>" + "(minX;maxX;minY;maxY)" + "</span><br>" + c.Xmin + ";" + c.Xmax + ";" + c.Ymin + ";" + c.Ymax);
                stg.AppendLine("<br><br>");
                stg.AppendLine("<B><span style =\"color: RGB(230, 20, 0); \">" + "Height: " + "</span></B>" + Math.Round(c.Height_InMeter / 1000, 2) + " km");
                stg.AppendLine("<br>");
                stg.AppendLine("<B><span style =\"color: RGB(230, 20, 0); \">" + "Top width: " + "</span></B>" + Math.Round(c.Width_Up_InMeter / 1000, 2) + " km");
                stg.AppendLine("<br>");
                stg.AppendLine("<B><span style =\"color: RGB(230, 20, 0); \">" + "Bottom width: " + "</span></B>" + Math.Round(c.Width_Down_InMeter / 1000, 2) + " km");

                placemk.Description = new SharpKml.Dom.Description
                {
                    Text = stg.ToString()
                };

                doc.AddFeature(placemk);
            }

            SharpKml.Engine.KmlFile kml = SharpKml.Engine.KmlFile.Create(doc, true);

            using (var stream = System.IO.File.OpenWrite(kml_filepath))
            {
                kml.Save(stream);
            }
        }

        public AFS2_Grid(int Level)
        {
            // Calcul basé sur la valeur brut de la la grille Aerofly FS2 qui se situe dans une espaces situé entre 0 et 65536

            // calcul de nombre de cellule contenue dans la largeur du globe terrestre en fonction du "Level" demandé
            this.Level = Level;
            this.nb_cellinEarthWidth = (int)Math.Pow(2, Level);
            double EarthValuMax = 65536.0;

            // cette fois ci la largeur et hauteur ne s'exprime plus en degré mais en unité décimal par rapport à la grille avec conversion
            this.Grid_Width = EarthValuMax / nb_cellinEarthWidth;

            // intiialisation du tableau qui contiendra l'ensembre des points correspondant aux coins inferieur gauche des cellules de la grille
            this.Ref_points = new PointD[nb_cellinEarthWidth + 1];

            for (int i = 0; i < nb_cellinEarthWidth + 1; i++)
            {
                double point = this.Grid_Width * i;
                double[] coord = LonLatFromWorldGrid(new double[2] { point, point });
                this.Ref_points[i] = new PointD(coord[0], coord[1]);
            }
        }

        public void FillCells_bis()
        {
            // On remplit toutes les cellules

            this.Cells = new List<Cellule>();

            for (int line = 0; line < nb_cellinEarthWidth; line++)
            {
                for (int col = 0; col < nb_cellinEarthWidth; col++)
                {
                    double width = this.Ref_points[col + 1].X - this.Ref_points[col].X;
                    double height = this.Ref_points[line + 1].Y - this.Ref_points[line].Y;

                    Cellule c = new Cellule(this.Ref_points[col].X, this.Ref_points[col].Y, width, height, true);

                    this.Cells.Add(c);
                }
            }
        }

        public enum SelectMode
        {
            Exclusive,
            Cross,
            Total
        }

        public void FillCells()
        {
            // On remplit toutes les cellules
            this.Cells = new List<Cellule>();

            for (int line = 0; line < nb_cellinEarthWidth; line++)
            {
                for (int col = 0; col < nb_cellinEarthWidth; col++)
                {
                    double xmin = this.Ref_points[col].X;
                    double xmax = this.Ref_points[col].X + this.Ref_points[col + 1].X;
                    double ymin = this.Ref_points[col].Y;
                    double ymax = this.Ref_points[col].Y + this.Ref_points[col + 1].Y;

                    Cellule c = new Cellule(xmin, xmax, ymin, ymax);

                    c.SetAFS2Name();

                    this.Cells.Add(c);
                }
            }
        }

        public void FillCells_INBBox(double BBox_xmin, double BBox_xmax, double BBox_ymin, double BBox_ymax, SelectMode Mode)
        {
            // On remplit toutes les cellules
            this.Cells = new List<Cellule>();

            for (int line = 0; line < nb_cellinEarthWidth; line++)
            {
                for (int col = 0; col < nb_cellinEarthWidth; col++)
                {
                    double xmin = this.Ref_points[col].X;
                    double xmax = this.Ref_points[col + 1].X;
                    double ymin = this.Ref_points[line].Y;
                    double ymax = this.Ref_points[line + 1].Y;

                    bool test_Xmax = false;
                    bool test_Xmin = false;
                    bool test_Ymax = false;
                    bool test_Ymin = false;

                    if ((xmax >= BBox_xmin) && (xmax <= BBox_xmax)) test_Xmax = true;
                    if ((xmin >= BBox_xmin) && (xmin <= BBox_xmax)) test_Xmin = true;
                    if ((ymax >= BBox_ymin) && (ymax <= BBox_ymax)) test_Ymax = true;
                    if ((ymin >= BBox_ymin) && (ymin <= BBox_ymax)) test_Ymin = true;

                    if (Mode == SelectMode.Total && ((!test_Xmax && !test_Xmin) || (!test_Ymax && !test_Ymin)))
                    {
                        // on inverse le test et on test si la bbox est dans la cellule
                        if ((BBox_xmax >= xmin) && (BBox_xmax <= xmax)) test_Xmax = true;
                        if ((BBox_xmin >= xmin) && (BBox_xmin <= xmax)) test_Xmin = true;
                        if ((BBox_ymax >= ymin) && (BBox_ymax <= ymax)) test_Ymax = true;
                        if ((BBox_ymin >= ymin) && (BBox_ymin <= ymax)) test_Ymin = true;
                    }

                    if (test_Ymin || test_Ymax)
                    {

                        string test = "";
                    }

                    if (Mode == SelectMode.Exclusive)
                    {
                        // Si la celllule est completement contenu totalement dans la bbox
                        if ((test_Xmax && test_Xmin && test_Ymax && test_Ymin))
                        {
                            Cellule c = new Cellule(xmin, xmax, ymin, ymax);
                            c.SetAFS2Name();
                            c.SetInMeter();
                            this.Cells.Add(c);
                        }
                    }

                    if (Mode == SelectMode.Cross || Mode == SelectMode.Total)
                    {
                        // Si une partie de la cellule intersect la bbox

                        if ((test_Xmax || test_Xmin) && (test_Ymax || test_Ymin))
                        {
                            Cellule c = new Cellule(xmin, xmax, ymin, ymax);
                            c.SetAFS2Name();
                            c.SetInMeter();
                            this.Cells.Add(c);
                        }
                    }

                }
            }
        }

        // test - creations des polygones formant la grille
        public SharpKml.Dom.Polygon GetPolygon(PointD point1, PointD point2)
        {

            // creations des points composants la GeoBox
            SharpKml.Base.Vector v1 = new SharpKml.Base.Vector(point1.Y, point1.X);
            SharpKml.Base.Vector v2 = new SharpKml.Base.Vector(point1.Y, point2.X);
            SharpKml.Base.Vector v3 = new SharpKml.Base.Vector(point2.Y, point2.X);
            SharpKml.Base.Vector v4 = new SharpKml.Base.Vector(point2.Y, point1.X);

            SharpKml.Dom.CoordinateCollection coords = new SharpKml.Dom.CoordinateCollection(new List<SharpKml.Base.Vector> { v1, v2, v3, v4, v1 });

            SharpKml.Dom.LinearRing lring = new SharpKml.Dom.LinearRing
            {
                Coordinates = coords
            };

            SharpKml.Dom.OuterBoundary outb = new SharpKml.Dom.OuterBoundary
            {
                LinearRing = lring
            };

            string name = GetWorldGridNAme_FromLonLat(point1);

            SharpKml.Dom.Polygon polygon = new SharpKml.Dom.Polygon
            {
                Id = name,
                Extrude = true,
                AltitudeMode = SharpKml.Dom.AltitudeMode.ClampToGround,
                OuterBoundary = outb
            };

            return polygon;
        }

        // test - creations des polygones formant la grille
        public SharpKml.Dom.Polygon Cell2Polygon(Cellule cell)
        {

            // creations des points composants la GeoBox
            SharpKml.Base.Vector v1 = new SharpKml.Base.Vector(cell.Ymin, cell.Xmin);
            SharpKml.Base.Vector v2 = new SharpKml.Base.Vector(cell.Ymax, cell.Xmin);
            SharpKml.Base.Vector v3 = new SharpKml.Base.Vector(cell.Ymax, cell.Xmax);
            SharpKml.Base.Vector v4 = new SharpKml.Base.Vector(cell.Ymin, cell.Xmax);

            SharpKml.Dom.CoordinateCollection coords = new SharpKml.Dom.CoordinateCollection(new List<SharpKml.Base.Vector> { v1, v2, v3, v4, v1 });

            SharpKml.Dom.LinearRing lring = new SharpKml.Dom.LinearRing
            {
                Coordinates = coords
            };

            SharpKml.Dom.OuterBoundary outb = new SharpKml.Dom.OuterBoundary
            {
                LinearRing = lring
            };

            SharpKml.Dom.Polygon polygon = new SharpKml.Dom.Polygon
            {
                Id = "Area",
                Extrude = true,
                AltitudeMode = SharpKml.Dom.AltitudeMode.ClampToGround,
                OuterBoundary = outb
            };

            return polygon;
        }

        private SharpKml.Dom.Placemark Polygon2placmk(SharpKml.Dom.Polygon polygon, string name, SharpKml.Dom.Style stl)
        {
            // Creation du conteneur du polygone
            SharpKml.Dom.Placemark placemk = new SharpKml.Dom.Placemark
            {
                Name = name,
                Geometry = polygon,
                StyleUrl = new Uri("#" + stl.Id, UriKind.Relative)

            };

            return placemk;
        }

        private SharpKml.Dom.Style CreateStyle(Color Fill_color, Color Line_Color, double epaisseur, byte alpha, int FileID, int IDnb)
        {
            SharpKml.Base.Color32 lineColor = new SharpKml.Base.Color32(Line_Color.A, Line_Color.B, Line_Color.G, Line_Color.R);
            SharpKml.Base.Color32 fillColor = new SharpKml.Base.Color32(alpha, Fill_color.B, Fill_color.G, Fill_color.R);
            double epaisseurTrait = epaisseur;

            // definition des couleurs de traits
            SharpKml.Dom.Style stl = new SharpKml.Dom.Style
            {
                Id = "stylePolygon_" + FileID + "_" + IDnb,
                Polygon = new SharpKml.Dom.PolygonStyle
                {
                    Color = fillColor,
                    Fill = true,
                    Outline = true
                },

                Line = new SharpKml.Dom.LineStyle
                {
                    Color = lineColor,
                    Width = epaisseurTrait
                }
            };

            return stl;
        }

        // *******************************************
        public double WorldGridConstantA = 2.3311223704144;
        public int TM_PI = 180;

        public string GetWorldGridNAme_FromLonLat(PointD point)
        {
            int nb_cellinEarthWidth = (int)Math.Pow(2, 16);

            double x = point.X / TM_PI;
            double y = point.Y / TM_PI;
            y = Math.Tan(WorldGridConstantA * y) / WorldGridConstantA;
            x = 0.5 + 0.5 * x;
            y = 0.5 + 0.5 * y;
            double[] coord = new double[2] { nb_cellinEarthWidth * x, nb_cellinEarthWidth * y };

            // puis convertir x et y en hexadecimal pour avoir l'ecriture correspondant au nom de fichier
            string name = Convert.ToInt32(coord[0]).ToString("X") + "_" + Convert.ToInt32(coord[1]).ToString("X");
            return name;
        }

        private double[] LonLatFromWorldGrid(double[] world_grid)
        {
            int nb_cellinEarthWidth = (int)Math.Pow(2, 16);

            double x = world_grid[0] / nb_cellinEarthWidth;
            double y = world_grid[1] / nb_cellinEarthWidth;

            x = 2.0 * (x - 0.5);
            y = 2.0 * (y - 0.5);
            y = Math.Atan(WorldGridConstantA * y) / WorldGridConstantA;
            x *= TM_PI;
            y *= TM_PI;
            return new double[2] { x, y };
        }


        // ***************************************
        // largeur en degré de longitude d'une cellule en fonction de son Level
        private double GetWidthCell_inDegre(int level)
        {
            // calcul de nombre de cellule contenue dans la largeur du globe terrestre en fonction du "Level" demandé
            int nb_cellinEarthWidth = (int)Math.Pow(2, level);

            // Comme la largeur du globe est exprimé sur 360° (+180, -180) -> longitude / x
            double width = 360.0 / nb_cellinEarthWidth;
            return width;
        }

        // hauteur en degré de latitude d'une cellule en fonction de son Level
        private double GetHeightCell_inDegre(int level)
        {
            // calcul de nombre de cellule contenue dans la largeur du globe terrestre en fonction du "Level" demandé
            int nb_cellinEarthHeight = (int)Math.Pow(2, level);

            // Comme la hauteur du globe est exprimé sur 180° (+90, -90) -> latitude / y
            double height = 175 / nb_cellinEarthHeight;
            return height;
        }


    }


    public static class Tools
    {

        // ***************************************************************************************
        // ******** Conversion Degré -> metre de la Latitude et Longitude **********************

        // En faisant l'approximation que la terre est une sphère de circonférence C = 40 000 km
        public static int CirconfTerre = 40000000;

        // La latitude se mesure le long des méridiens qui sont tous des demi-grands cercles. Ils ont donc tous la même longueur. 
        static double Latitude_InMeter = 40000000 / 360.0;

        // La longitude se mesure le long des parallèles qui sont des cercles de diametre variable fonction de la latitude (plus la latitude est élevée (proche de 90°), plus le cercle est petit)
        private static double Longitude_inMeter(double latitude)
        {
            // calcul de la largeur en metre pour 1° de longitude
            double angleRadian = latitude * (Math.PI / 180.0);
            double result = CirconfTerre * Math.Cos(angleRadian) / 360.0;
            return result;
        }

        public static double degre_lat_from_meter(double lenght)
        {
            double latitude = lenght / (CirconfTerre / 360.0);
            return latitude;
        }

        public static double degre_long_from_meter(double lenght, double latitude)
        {
            double angleRadian = latitude * (Math.PI / 180.0);
            double longitude = lenght / (CirconfTerre * Math.Cos(angleRadian) / 360.0);
            return longitude;
        }

        public static PointD GetPointB_xmeter_from_PointA(PointD A, double lenghtInmeter, Vecteur3D direction)
        {
            double y_offset = direction.Y * lenghtInmeter;
            double x_offset = direction.X * lenghtInmeter;

            double latitude_offset = degre_lat_from_meter(y_offset);
            double longitude_offset = degre_long_from_meter(x_offset, A.Y);

            return new PointD(A.X + longitude_offset, A.Y + latitude_offset);
        }


    }





}
